#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <stdlib.h>
#include <nwtypes.h>
#include <portable.h>
#include <nwsync.h>
#include <ncpext.h>
#include <advanced.h>
#include <process.h>
#include <nwmisc.h>
#include <errno.h>
#include <ctype.h>
#include <io.h>
#include <direct.h>
#include <fortune.h>

long semHandle;
WORD openCount;

int main(int argc, char **argv)
{
   int ccode = 0;
   char arg[128];
   char *pwd;

   /* Save off current thread group id */
   myThreadGroupID = GetThreadGroupID();

   strcpy(arg,argv[0]);

   /* Find the .ext and remove the extension */
   if ( (pwd = strrchr(arg,'.')) != NULL)
      *++pwd = '\0';

   strcat(arg,FORTUNE_EXT);

   /* Set defaults for file and ncp extension name */
   FortuneFile = arg;
   NCPExtension = NCP_EXTENSION_NAME;

   getopts(argc,argv);

   ccode = OpenFortuneFile(FortuneFile);
   if (!ccode)
   {
      printf("Using Fortune File %s\n\n",FortuneFile);

      /* allocate semaphore with count of 1 */
      ccode = OpenSemaphore(NCPExtension,1,&semHandle,&openCount);

      if (ccode)
      {
         printf("Error %d, unable to open semaphore %s\n\n",ccode, NCPExtension);
         exit(ccode);
      }

      /* register our ncp extension, with FortuneHandler() */
      ccode = NWRegisterNCPExtension(NCPExtension, FortuneHandler,
                                    NULL,NULL,1,0,0,&queryData);

      if (!ccode)
      {
         /* FortuneUnload called when nlm is unloaded */
         AtUnload(FortuneUnload);
         /* Exit our thread, leave code in memory  */
         ExitThread(TSR_THREAD,0);
      }
      else
         printf("Error %d, unable to register NCP Extension %s\n\n",ccode, NCPExtension);
   }
   else
      printf("Error %d, unable to open %s\n",ccode,FortuneFile);

   return (ccode);
}

BYTE FortuneHandler(NCPExtensionClient *client,void* reqData,
                 LONG reqDataLen, void* replyData,LONG* replyDataLen)
{
   int ccode;
   char fortuneBuffer[MAX_FORTUNE_LEN];
   int fortuneBufferSize = MAX_FORTUNE_LEN;
   int savedThreadGroupID = GetThreadGroupID();

   SetThreadGroupID(myThreadGroupID);
   
   /* Removed unused parameter warning, reData
      contains the request data from the client if needed */
   reqData = reqData;

   if (*replyDataLen < reqDataLen)
      reqDataLen = *replyDataLen;

   /* print out calling task's connection and task */
   printf("NCP Extension Request from connection %d task %d\n",client->connection,client->task);

   /* Wait on the semaphore, blocks other threads after the first */
   WaitOnSemaphore(semHandle,18);

   /* Read the fortune into fortuneBuffer, then copy the
      fortune into the replyData buffer */
   ccode = ReadNextFortune(FortuneFile, fortuneBuffer, &fortuneBufferSize);
   if (!ccode)
      strcpy((char*)replyData,fortuneBuffer);
   else
      ((char *)replyData)[0] = '\0';

   /* Set length of reply buffer for OS */
   *replyDataLen = fortuneBufferSize;

   SignalSemaphore(semHandle);

   /* reset thread group ID */
   SetThreadGroupID(savedThreadGroupID);
   return(0);
}


void FortuneUnload()
{
   /* Called at unload time,
      clean up allocated resources */

   CloseSemaphore(semHandle);
   NWDeRegisterNCPExtension(queryData);
}

int OpenFortuneFile(char * fileName)
{
   // check for existence of file
   return ( access(fileName,0) != 0);
}


int ReadNextFortune(char * fileName, char * buffer, int * size)
{
   int ccode = 1;
   static long fortuneOffset = 0;
   FILE * fortuneFP;
   char fortune[MAX_FORTUNE_LEN+1];

   /* read the fortune file using ordinary stdio.h type
      functions fopen(), fseek(), fclose() */
   fortuneFP = fopen(fileName,"r");
   if (fortuneFP)
   {
      fseek(fortuneFP,fortuneOffset,SEEK_SET);
      if (!ReadFortune(fortuneFP,fortune,MAX_FORTUNE_LEN,&fortuneOffset))
      {
         strcpy(buffer,fortune);
         // include '\0' at end of buffer
         *size = strlen(buffer) + 1;
         fclose(fortuneFP);
         ccode = 0;
      }
   }
   return ccode;
}

int ReadFortune(FILE * fortuneFP, char * buffer, int max, long *offset)
{
   int ccode = 1;
   int ch;
   int i = 0;

   /* Fortunes are delimited by '%', read characters until
      next '%' or EOF */
   ch = getc(fortuneFP);
   if (ch == '%')
   {
      while ( (ch = getc(fortuneFP)) != '%' && ch != EOF)
         if (i < max - 1)
            buffer[i++] = ch;

      buffer[i] = '\0';
      if (ch == '%')
      {
         ungetc(ch,fortuneFP);
         *offset = ftell(fortuneFP);
      }
      else
      {
         *offset = 0;
      }

      ccode = 0;
   }
   return ccode;
}

void getopts(int argc, char **argv)
{
   int    i;
   char * arg;

   /* parse command line parameters */
   for(i= 1; i < argc; i++)
   {
      arg = argv[i];
      if ( *arg++ == '-')
      switch ( tolower(*arg++) )
      {
         case 'f':
            /* -f<fortune-file> sets FortuneFile to new name */
            FortuneFile = arg;
            break;

         case 'n':
            /* -n<NCPExtension> sets NCPExtension to new name */
            NCPExtension = arg;
            break;
      }
   }
}
